/* ----------------------------------------------------------------------- */
/*         Fernando Berzal Galiano       ETS Ingeniera Informtica        */
/* ----------------------------------------------------------------------- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "DB.H"
#include "DB_DAT.H"
#include "DB_FRM.H"

#include "FBG_DATE.H"

/* ----------------------------------------------------------------------- */
/*            Gestin del fichero principal de la base de datos            */
/* ----------------------------------------------------------------------- */

static void *CargaDato (DataBase BD, long numero)
{
  void *dato = NULL;
  FILE *fich = fopen(BD->fichDatos,"rb");

  if (fich) {

     fseek (fich, (numero-1)*BD->dim, SEEK_SET);

     dato=malloc(BD->dim);

     if (dato)
	fread (dato, BD->dim, 1, fich);

     fclose (fich);
  }

  return dato;
}

/* ----------------------------------------------------------------------- */

static void GuardaDato (DataBase BD, long numero, void *dato)
{
  FILE *fich=fopen(BD->fichDatos,"r+b");

  if (fich) {
     fseek (fich, (numero-1)*BD->dim, SEEK_SET);
     fwrite (dato, BD->dim, 1, fich);
     fclose (fich);
  }
}

/* ----------------------------------------------------------------------- */

static void BorraDato (DataBase BD, long numero)
{
  long registro;
  long longitud;
  FILE *fich   ;
  FILE *aux    ;
  void *dato   = malloc(BD->dim);

  if (dato) {

     fich = fopen(BD->fichDatos,"rb");
     aux  = fopen("bd.tmp","wb");

     if (fich && aux) {

	fseek (fich, 0, SEEK_END);
	longitud = ftell (fich);
	fseek (fich, 0, SEEK_SET);

	for (registro=1; ftell(fich)<longitud; registro++) {

	    fread (dato, BD->dim, 1, fich);

	    if (registro!=numero)
	       fwrite (dato, BD->dim, 1, aux);
	}

	fclose (fich);
	fclose (aux);
	remove (BD->fichDatos);
	rename ("bd.tmp",BD->fichDatos);

     } else {

	if (fich) fclose (fich);
	if (aux)  fclose (aux);

     }

     free (dato);
  }

}

/* ----------------------------------------------------------------------- */
/*                         Gestin de campos MEMO                          */
/* ----------------------------------------------------------------------- */

#define MARCA_MEMO ""

/* ----------------------------------------------------------------------- */

static void LeeCad (FILE *fichero, unsigned char *cadena, int longitud)
{
  fgets (cadena, longitud, fichero);

  if (cadena[strlen(cadena)-1]=='\n')
     cadena[strlen(cadena)-1] = '\0';
  else
     while (fgetc(fichero)!='\n')
	   ;
}

/* ----------------------------------------------------------------------- */
/*  Carga el contenido del campo MEMO de la ficha 'numero'                 */

static Lista CargaMEMO (DataBase BD, int campo, long numero)
{
  FILE *f;
  long  posicion;
  char  cadena[CAMPOMEMO];
  char *s;
  Lista l = CreaLista(0);

  if (l) {

     f = fopen(IndiceCampoMEMO(BD,campo),"rb");

     if (f) {

	fseek (f, (numero-1)*sizeof(long), SEEK_SET);
	fread (&posicion, sizeof(long), 1, f);
	fclose (f);

	f = fopen(FicheroCampoMEMO(BD,campo),"rb");

	if (f) {

	   fseek (f, posicion, SEEK_SET);

	   do {
	       LeeCad (f, cadena, CAMPOMEMO);

	       if (strcmp(cadena,MARCA_MEMO)) {
		  s = (char *) malloc (strlen(cadena)+1);
		  strcpy (s, cadena);
		  InsertaNodoLista (l, FinLista(l), s);
	       }

	   } while (strcmp(cadena,MARCA_MEMO));

	   fclose (f);

	}

     }

  }

  return l;
}

/* ----------------------------------------------------------------------- */
/*  Guarda fsicamente el contenido de un campo MEMO                       */

static void GuardaMEMO_Final (DataBase BD, int campo, Lista memo)
{
  FILE *fNDX;
  FILE *fMEM;
  long  posicion;
  Nodo  nodo;

  fMEM=fopen(FicheroCampoMEMO(BD,campo),"a+b");

  if (fMEM) {

     fseek (fMEM, 0, SEEK_END);
     posicion = ftell(fMEM);

     nodo = InicioLista (memo);

     while (nodo!=FinLista(memo)) {
	 fprintf(fMEM, "%s\n", LeeNodoLista(memo,nodo));
	 nodo = SiguienteLista(memo, nodo);
     }

     fprintf (fMEM, MARCA_MEMO);
     fprintf (fMEM, "\n");
     fclose  (fMEM);

     fNDX=fopen(IndiceCampoMEMO(BD,campo),"a+b");
     fwrite (&posicion, sizeof(long), 1, fNDX);
     fclose (fNDX);
  }
}

  /* --- */

static void GuardaMEMO_Interno (DataBase BD, int numero, int campo, Lista memo)
{
  long  ficha;
  long  fichas;
  long  posicion;
  char  cadena[CAMPOMEMO];
  Nodo  nodo;
  FILE *fNDX = fopen( IndiceCampoMEMO(BD,campo),"r+b");
  FILE *fMEM = fopen(FicheroCampoMEMO(BD,campo),"rb");
  FILE *faux = fopen("memo.tmp", "wb");;

  fseek (fNDX, 0, SEEK_END);
  fichas = ftell(fNDX) / sizeof(long);


  for (ficha=1; ficha<=fichas; ficha++) {

    if (ficha<numero) {

       do {
	   LeeCad (fMEM, cadena, CAMPOMEMO);
	   fprintf(faux, "%s\n", cadena);
       } while (strcmp(cadena,MARCA_MEMO));

    } else if (ficha==numero) {

       nodo = InicioLista (memo);

       while (nodo!=FinLista(memo)) {
	     fprintf(faux, "%s\n", LeeNodoLista(memo,nodo));
	     nodo = SiguienteLista(memo, nodo);
       }

       fprintf (faux, MARCA_MEMO);
       fprintf (faux, "\n");

       fseek (fNDX, numero*sizeof(long), SEEK_SET);
       fread (&posicion, sizeof(long), 1, fNDX );
       fseek (fNDX, numero*sizeof(long), SEEK_SET);
       fseek (fMEM, posicion,            SEEK_SET);

    } else {

       posicion = ftell (faux);

       fwrite (&posicion, sizeof(long), 1, fNDX);

       do {
	   LeeCad (fMEM, cadena, CAMPOMEMO);
	   fprintf(faux, "%s\n", cadena);
       } while (strcmp(cadena,MARCA_MEMO));

    }

  }

  fclose (faux);
  fclose (fMEM);
  fclose (fNDX);

  remove (FicheroCampoMEMO(BD,campo));
  rename ("memo.tmp", FicheroCampoMEMO(BD,campo));
}

  /* --- */

static void GuardaMEMO (DataBase BD, int numero, int campo, Lista memo)
{
  if (numero==0)
     GuardaMEMO_Final (BD, campo, memo);
  else
     GuardaMEMO_Interno (BD, numero, campo, memo);
}

/* ----------------------------------------------------------------------- */
/*  Libera la memoria ocupada por los datos de un campo MEMO               */

static void LiberaMEMO (Lista memo)
{
 Nodo  nodo = InicioLista (memo);

 while (nodo!=FinLista(memo)) {
       free (LeeNodoLista(memo,nodo));
       nodo = SiguienteLista(memo,nodo);
 }

 DestruyeLista (&memo);
}

/* ----------------------------------------------------------------------- */
/*  Elimina fsicamente el campo memo de la ficha 'numero'                 */

static void BorraMEMO (DataBase BD, int campo, long numero)
{
  FILE *fNDX;
  FILE *fMEM;
  FILE *faux;

  char s[CAMPOMEMO];
  long posicion;
  long longitud;
  long ficha;
  long fichas;

  fNDX = fopen(IndiceCampoMEMO(BD,campo),"rb");
  fMEM = fopen(FicheroCampoMEMO(BD,campo),"rb");

  fseek (fNDX, 0, SEEK_END);
  fichas = ftell(fNDX)/sizeof(long);
  fseek (fNDX, (numero-1)*sizeof(long), SEEK_SET);
  fread (&posicion, sizeof(long), 1, fNDX);

  /* Actualizacin fichero MEMO */

  faux = fopen("mem.tmp","wb");

  ficha = 1;

  while (ficha<=fichas) {
	LeeCad (fMEM, s, CAMPOMEMO);

	if (ficha!=numero)
	   fprintf (faux, "%s\n", s);

	if (!strcmp(s,MARCA_MEMO)) {

	   if (ficha==numero)
	      longitud = ftell(fMEM) - posicion;

	   ficha++;
	}
  }

  fclose (fMEM);
  fclose (faux);
  remove (FicheroCampoMEMO(BD,campo));
  rename ("mem.tmp", FicheroCampoMEMO(BD,campo));

  /* Actualizacin fichero ndice */

  faux = fopen ("ndx.tmp","wb");
  fseek (fNDX, 0, SEEK_SET);

  for ( ficha=1; ficha<=fichas; ficha++) {

      fread (&posicion, sizeof(long), 1, fNDX);

      if (ficha>numero)
	 posicion -= longitud;

      if (ficha!=numero)
	 fwrite (&posicion, sizeof(long), 1, faux);
  }

  fclose (fNDX);
  fclose (faux);
  remove (IndiceCampoMEMO(BD,campo));
  rename ("ndx.tmp", IndiceCampoMEMO(BD,campo));
}

/* ----------------------------------------------------------------------- */
/*  Acceso a los campos de las fichas                                      */
/* ----------------------------------------------------------------------- */

void *Campo (Ficha ficha, int posicion)
{
  if (posicion>0 && posicion<=CamposBD(ficha->BD))
     return ficha->info[posicion-1];
  else
     return NULL;
}

/* ----------------------------------------------------------------------- */

void CadEnum (DataBase BD, int campo, int posicion, char *cadena)
{
  Lista valores = ValoresCampoEnumerado (BD,campo);
  Nodo  nodo    = InicioLista (valores);

  if ((posicion>0)&&(posicion<=CardinalCampoEnumerado(BD,campo))) {

     while (posicion>1) {
	   nodo = SiguienteLista (valores, nodo);
	   posicion--;
     }

     strcpy (cadena,(char*)LeeNodoLista(valores,nodo));

  } else

     strcpy (cadena,"");
}

/* ----------------------------------------------------------------------- */

void CadRef (DataBase BD, int campo, long valor, char *cadena)
{
  DataBase RefBD;
  Ficha    RefFicha;

  if (valor>0) {
     RefBD    = CargaBD    ( BDCampoReferencia (BD, campo) );
     RefFicha = CargaFicha ( RefBD, valor);
     CadenaCampo (RefFicha,PKCampoReferencia(BD,campo),cadena);
     LiberaFicha (&RefFicha);
     LiberaBD    (&RefBD);
  } else
     strcpy ( cadena, "" );
}

/* ----------------------------------------------------------------------- */

void CadenaCampo (Ficha ficha, int campo, char *cadena)
{
  switch (Tipo_Campo(ficha->BD, campo)) {

	 case Entero:     ltoa ( *(long *)Campo(ficha,campo), cadena, 10);
			  break;

	 case Real:       gcvt ( *(double *)Campo(ficha,campo), 10, cadena);
			  break;

	 case Cadena:     strcpy (cadena, (char *)Campo(ficha,campo));
			  break;

	 case Fecha:      CadenaFecha (cadena, *(long *)Campo(ficha,campo));
			  break;

	 case Enumerado:  CadEnum (ficha->BD, campo, *(short *)Campo(ficha,campo), cadena );
			  break;

	 case MEMO:       strcpy (cadena, "MEMO" );
			  break;

	 case Referencia: CadRef (ficha->BD, campo, *(long *)Campo(ficha,campo), cadena );
			  break;
  }
}


/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*                     Operaciones bsicas con fichas                      */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

long NumeroFicha (Ficha f)
{
  return (f->numero);
}

DataBase BDFicha (Ficha f)
{
  return (f->BD);
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Ficha CargaFicha (DataBase BD, long numero)
{
  Ficha f        =  NULL;
  int   campo    ;
  int   posicion = 0;
  void *dato     ;

  if (numero>0 && numero<=FichasBD(BD)) {

     dato = CargaDato (BD, numero);

     if (dato) {

	f = malloc (sizeof(struct Ficha));

	if (f) {

	   f->BD     = BD;
	   f->numero = numero;
	   f->dato   = dato;
	   f->info   = calloc(BD->NumCampos,sizeof(void*));

	   if (f->info) {

	      for (campo=1; campo<=BD->NumCampos; campo++) {

		  switch (Tipo_Campo(BD,campo)) {

		    case Fecha:
		    case Referencia:
		    case Entero:     f->info[campo-1] = (char *)dato + posicion;
				     posicion += sizeof (long int);
				     break;

		    case Real:       f->info[campo-1] = (char *)dato + posicion;
				     posicion += sizeof (double);
				     break;

		    case Cadena:     f->info[campo-1] = (char *)dato + posicion;
				     posicion += LongitudCampoCadena(BD,campo);
				     break;

		    case Enumerado:  f->info[campo-1] = (char *)dato + posicion;
				     posicion += sizeof (short int);
				     break;

		    case MEMO:       f->info[campo-1] = CargaMEMO(BD,campo,numero);
				     break;
		  }
	      }

	   } else {
	      free (dato);
	      free (f);
	      f = NULL;
	   }

	} else free(dato);
     }
  }

  return f;
}

/* ----------------------------------------------------------------------- */

void GuardaFicha (Ficha f)
{
  int  campo;

  /* Campos MEMO */

  for (campo=1; campo<=f->BD->NumCampos; campo++)
      if (Tipo_Campo(f->BD,campo)==MEMO)
	 GuardaMEMO (f->BD, f->numero, campo, f->info[campo-1]);

  /* Registro del fichero principal */

  if (f->numero==0)                     /* Ficha no guardada anteriormente */
     f->numero = FichasBD(f->BD)+1;

  GuardaDato (f->BD, f->numero, f->dato);
}

/* ----------------------------------------------------------------------- */
/*  Libera el espacio ocupado en memoria por una ficha                     */

void LiberaFicha (Ficha *f)
{
  int  campo;

  for (campo=1; campo<=(*f)->BD->NumCampos; campo++) {

      switch (Tipo_Campo((*f)->BD,campo)) {

	case Fecha:
	case Entero:
	case Real:
	case Cadena:
	case Enumerado:
	case Referencia: break;

	case MEMO:       LiberaMEMO ( (*f)->info[campo-1] );
			 break;
      }

  }

  free ((*f)->info);
  free ((*f)->dato);
  free (*f);
  *f = NULL;
}

/* ----------------------------------------------------------------------- */
/*  Elimina fsicamente la ficha f, tambin libera la memoria que ocupa    */

void BorraFicha (Ficha *f)
{
  int   campo;

  if ((*f)->numero) { /* Ficha existente en disco */

     for (campo=1; campo<=(*f)->BD->NumCampos; campo++) {

      switch (Tipo_Campo((*f)->BD,campo)) {

	case Fecha:
	case Entero:
	case Real:
	case Cadena:
	case Enumerado:
	case Referencia: break;

	case MEMO:       BorraMEMO  ( (*f)->BD, campo, (*f)->numero);
			 LiberaMEMO ( (*f)->info[campo-1] );
			 break;
      }
     }

     BorraDato ((*f)->BD, (*f)->numero);
     free ((*f)->info);
     free ((*f)->dato);
     free (*f);
     *f = NULL;

  } else
     LiberaFicha (f);
}

/* ----------------------------------------------------------------------- */
/*  Crea una ficha                                                         */

Ficha CreaFicha (DataBase BD)
{
  int   campo;
  void *dato;
  Ficha f = malloc(sizeof(struct Ficha));

  if (f) {

	f->numero = 0;
	f->BD     = BD;
	f->dato   = malloc (BD->dim);
	f->info   = calloc (BD->NumCampos,sizeof(void*));

	if (f->dato && f->info) {

	   memset(f->dato,0,BD->dim);

	   for (campo=1; campo<=BD->NumCampos; campo++) {

	       dato = (char*)f->dato + OffsetCampo(f->BD,campo);

	       switch (Tipo_Campo(f->BD,campo)) {

		 case Fecha:      f->info[campo-1] = dato;
				  *(long *)dato = Hoy();
				  if (  *(long *)dato<MinCampoFecha(BD,campo)
				     || *(long *)dato>MaxCampoFecha(BD,campo) )
				     *(long *)dato = MinCampoFecha(BD,campo);
				  break;

		 case Entero:     f->info[campo-1] = dato;
				  *(long *)dato = MinCampoEntero(BD,campo);
				  break;

		 case Real:       f->info[campo-1] = dato;
				  *(double *)dato = MinCampoReal(BD,campo);
				  break;

		 case Cadena:     f->info[campo-1] = dato;
				  strcpy ( (char*)dato, "");
				  break;

		 case Enumerado:  f->info[campo-1] = dato;
				  *(short int *)dato = 1;
				  break;

		 case MEMO:       f->info[campo-1] = CreaLista (0);
				  break;

		 case Referencia: f->info[campo-1] = dato;
				  *(long *)dato = 0; /* Valor nulo */
				  break;
	       }
	   }

	} else {
	   if (f->dato) free (f->dato);
	   if (f->info) free (f->info);
	   free (f);
	   f = NULL;
	}
  }

  return f;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
