/* ----------------------------------------------------------------------- */
/*         Fernando Berzal Galiano       ETS Ingeniera Informtica        */
/* ----------------------------------------------------------------------- */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include "FBG_CRT.H"
#include "FBG_KBD.H"
#include "FBG_MENU.H"
#include "FBG_UTIL.H"
#include "FBG_TEXT.H"
#include "FBG_DATE.H"

#include "DB.H"
#include "DB_DAT.H"
#include "DB_SRC.H"
#include "DB_FRM.H"
#include "DB_DOS.H"

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

#define LINEA       80+1

#define CABECERA      "ͻ"
#define INTERMEDIO    "                                                                             "
#define SEPARADOR     "Ķ"
#define PIE           "ͼ"
#define LINEA_VACIA   "                                                                                "

/* ----------------------------------------------------------------------- */
/*        Entrada/Salida por pantalla de fichas de una base de datos       */
/* ----------------------------------------------------------------------- */

/* ----------------------------------------------------------------------- */
/*  Mostrar fichas por pantalla                                            */
/* ----------------------------------------------------------------------- */

typedef struct Pantalla {
			  int longitud;
			  Lista info;
			};

typedef struct Pantalla *Pantalla;

/* ----------------------------------------------------------------------- */

Pantalla CreaPantalla (void)
{
  Pantalla p = malloc (sizeof(Pantalla));

  p->info     = CreaLista (LINEA);
  p->longitud = 0;

  return p;
}

/* ----------------------------------------------------------------------- */

void DestruyePantalla (Pantalla *p)
{
 DestruyeLista (&((*p)->info));
 free (*p);
 *p = NULL;
}          

/* ----------------------------------------------------------------------- */

void AgregaLineaPantalla (Pantalla pantalla, char *linea)
{
  InsertaNodoLista (pantalla->info, FinLista(pantalla->info), linea);
  pantalla->longitud++;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Pantalla Plantilla (Ficha ficha)
{
  int      i;
  int      campo;

  char     cadena[LINEA];
  char     aux[LINEA];

  Linea    lMEMO;
  Texto    tMEMO;

  Pantalla scr = CreaPantalla();

  /* Creacin de la plantilla de la ficha */

  AgregaLineaPantalla (scr, CABECERA);

  for (campo=1; campo<=CamposBD(ficha->BD); campo++) {

    strcpy (cadena, INTERMEDIO);

    /* Ttulo del campo */

    strncpy (aux, NombreCampo(ficha->BD, campo), CAMPO);
    aux[CAMPO-1] = '\0';

    for (i=0; i<strlen(aux); i++)
	cadena[i+2] = aux[i];

    /* Informacin mostrada segn el tipo de campo */

    if (Tipo_Campo(ficha->BD,campo)!=MEMO) {

	CadenaCampo(ficha, campo, aux);

	for (i=0; i<strlen(aux); i++)
	    cadena[i+17] = aux[i];

    } else {         /* Campo MEMO */

	tMEMO = (Texto)Campo(ficha,campo);
	lMEMO = PrimeraLinea(tMEMO);

	while (lMEMO!=FinTexto(tMEMO)) {

	      strcpy (aux, CadenaLinea(tMEMO, lMEMO));

	      for (i=0; i<strlen(aux) && i<61; i++)
		  cadena[i+17] = aux[i];

	      if (lMEMO!=UltimaLinea(tMEMO)) {
		 AgregaLineaPantalla (scr, cadena);
		 strcpy (cadena, INTERMEDIO);
	      }

	      lMEMO = SiguienteLinea(tMEMO, lMEMO);
	}
    }

    AgregaLineaPantalla (scr, cadena);

    if (campo<CamposBD(ficha->BD))
       AgregaLineaPantalla (scr, SEPARADOR);
  }

  AgregaLineaPantalla (scr, PIE);

  return scr;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

void MuestraLinea (char *linea)
{
  int i=0;

  while (linea[i]!='\0') {

    if (strchr("ѻŶϼ",linea[i]))
       TextColor (CYAN);

    PutChar (linea[i]);

    if (strchr("ѻŶϼ",linea[i]))
       TextColor (GRIS);

    i++;
  }
}

/* ----------------------------------------------------------------------- */

int MuestraFicha ( Ficha ficha, CommandLine Comandos)
{
  int      tecla;
  int      i = 1;

  Nodo     inicio;
  Nodo     fin;
  Nodo     nodoScr;

  Pantalla scr = Plantilla (ficha);

  /* Visualizacin de la ficha */

  nodoScr = inicio = InicioLista (scr->info);
  Cursor (OCULTO);
  MuestraCL (Comandos, 18);

  GotoXY (1, 1);

  while ( (i<17) && (nodoScr!=FinLista(scr->info)) ) {
      MuestraLinea ((char *)LeeNodoLista(scr->info,nodoScr));
      fin     = nodoScr;
      nodoScr = SiguienteLista (scr->info, nodoScr);
      i++;
  }

  while (i<17) {
      MuestraLinea (LINEA_VACIA);
      i++;
  }

  do {  /* Lectura de la tecla pulsada */

      tecla = GetChar();

      if (!tecla) {       /* Cdigos de control 0 */

	 tecla =GetChar();

	 switch (tecla) {

	   case '\x48': /* Arriba */
			if (inicio!=InicioLista(scr->info)) {
			   inicio = AnteriorLista (scr->info, inicio);
			   fin = AnteriorLista (scr->info, fin);
			   Ventana(1,6,80,21);
			   ScrollDown (1);
			   Ventana(1,6,80,23);
			   GotoXY (1, 1);
			   MuestraLinea ((char*)LeeNodoLista(scr->info,inicio));
			}
			break;

	   case '\x50': /* Abajo  */
			if (fin!=UltimoLista(scr->info)) {
			   fin = SiguienteLista (scr->info, fin);
			   inicio = SiguienteLista (scr->info, inicio);
			   Ventana(1,6,80,21);
			   ScrollUp (1);
			   Ventana(1,6,80,23);
			   GotoXY (1, 16);
			   MuestraLinea ((char*)LeeNodoLista(scr->info, fin));
			}
			break;

	 }

	 tecla = 0;
      }

  } while (!TeclaCL(Comandos,tecla)&&(tecla!=ESCAPE));

  DestruyePantalla(&scr);
  Cursor (VISIBLE);
  return tecla;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  Edicin de fichas                                                      */
/* ----------------------------------------------------------------------- */

#define MAX(a,b) (((a)>(b))?(a):(b))

int EditaEnumerado (DataBase BD, int campo, short *Valor)
{
  int   tecla = 0;
  short ValorEnumerado = *Valor;

  BYTE x,y;

  int  i;
  int  longitud;
  char cadena[LINEA];

  unsigned cursor = EstadoCursor();

  PosicionCursor(&x,&y);

  Cursor (OCULTO);
  TextColor (BLANCO);

  for (i=1, longitud=0; i<=CardinalCampoEnumerado(BD,campo); i++) {
      CadEnum(BD,campo,i,cadena);
      longitud = MAX (longitud, strlen(cadena));
  }

  do {
      GotoXY (x, y);
      CadEnum (BD, campo, ValorEnumerado, cadena);
      PrintF ("%s", cadena);

      for (i=strlen(cadena); i<longitud; i++)
	  PutChar (' ');

      tecla = GetChar();

      if (!tecla) {                                /* Cdigos de control 0 */

	 tecla = GetChar();

	 switch (tecla) {

	   case '\x4B': /* Izquierda */
			if (ValorEnumerado>1)
			   ValorEnumerado--;
			else
			   ValorEnumerado = CardinalCampoEnumerado(BD,campo);
			break;

	   case '\x4D': /* Derecha */
			if (ValorEnumerado<CardinalCampoEnumerado(BD,campo))
			   ValorEnumerado++;
			else
			   ValorEnumerado = 1;
			break;

	   case '\x48': /* Arriba */
			ValorEnumerado = 1;
			break;

	   case '\x50': /* Abajo */
			ValorEnumerado = CardinalCampoEnumerado(BD,campo);
			break;

	 }
	 tecla = 0;

      }

  } while (tecla!=RETURN && tecla!=ESCAPE);

  TextColor (GRIS);
  GotoXY (x, y);
  CadEnum (BD, campo, ValorEnumerado, cadena);
  PrintF ("%s", cadena);

  for (i=strlen(cadena); i<longitud; i++)
      PutChar (' ');

  Cursor (cursor);

  if (tecla!=ESCAPE)
     *Valor = ValorEnumerado;

  return tecla;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  Editor de referencias                                                  */

void MuestraMarco (void)
{
  int i;

  TextColor (CYAN);
  GotoXY (9,3);
  PrintF ("ͻ");

  for (i=4; i<14; i++) {
     GotoXY ( 9, i);
     PrintF ("                                                                ");
  }

  GotoXY (9,14);
  PrintF ("ͼ");
  TextColor (GRIS);
}

  /* --- */

void ShowLinea ( char *cadena, int posicion, int inverso)
{
  if (inverso) {
     TextColor (NEGRO);
     TextBackground (GRIS);
  }

  GotoXY (1, posicion);
  PrintF ("%-62s", cadena);

  if (inverso) {
     TextColor (GRIS);
     TextBackground (NEGRO);
  }
}

  /* --- */

void ShowRef ( DataBase RefBD, int campo, long ficha, long inicio, int inverso)
{
  char  cadena[80];
  Ficha f = CargaFicha (RefBD, ficha);

  CadenaCampo (f, campo, cadena);

  ShowLinea (cadena, ficha-inicio+1, inverso);

  LiberaFicha (&f);
}

  /* --- */

int EditaReferencia (DataBase BD, int campo, long *valor, CommandLine CL)
{
  int      tecla    = ESCAPE;
  int      pk       = PKCampoReferencia (BD,campo);
  long     activo   = (*valor>0)? *valor: 1;
  long     inicio   = activo;
  DataBase RefBD    = CargaBD ( BDCampoReferencia (BD, campo) );

  unsigned cursor   = EstadoCursor();
  long     fichas   = FichasBD (RefBD);
  long     aux      ;

  Cursor (OCULTO);
  Presentacion ( NombreCampo (BD,campo) );
  MuestraMarco ();
  MuestraCL (CL, 16);
  Ventana(11,9,73,18);

  for (aux=inicio; aux<inicio+10 && aux<=fichas; aux++)
      ShowRef ( RefBD, pk, aux, inicio, (aux==activo));

  do {

     tecla = GetChar();

     if (!tecla) {                                   /* Cdigos de control */

	tecla = GetChar();

	switch (tecla) {

	       case '\x48': /* Arriba */
			    if (activo>1) {
			       ShowRef (RefBD, pk, activo, inicio, 0);
			       activo--;

			       if (activo<inicio) {
				  ScrollDown (1);
				  inicio--;
			       }

			       ShowRef (RefBD, pk, activo, inicio, 1);
			    }
			    break;

	       case '\x50': /* Abajo  */
			    if (activo<fichas) {
			       ShowRef (RefBD, pk, activo, inicio, 0);
			       activo++;

			       if (activo-inicio>9) {
				  ScrollUp (1);
				  inicio++;
			       }

			       ShowRef (RefBD, pk, activo, inicio, 1);
			    }
			    break;
	}

	tecla = '\0';
     }

  } while (tecla!=ESCAPE && tecla!=RETURN && !TeclaCL(CL,tecla) );


  if (tecla!=ESCAPE)
     *valor = activo;
  else
     *valor = 0;

  LiberaBD (&RefBD);

  Ventana(1,6,80,23);
  Cursor (cursor);

  return tecla;
}

/* --- */

int BuscaReferencia (DataBase BD, int campo, long *valor, char *tema)
{
  int      pk = PKCampoReferencia (BD,campo);
  int      tecla;
  int      relativa;
  long     ficha;
  Nodo     activo;
  Nodo     aux;
  DataBase RefBD;
  Busqueda search;
  unsigned cursor = EstadoCursor();

  Cursor (OCULTO);
  Presentacion ( NombreCampo (BD,campo) );

     MuestraMarco ();
     GotoXY (29,16);
     TextColor (AZUL_CLARO);
     PrintF ("Resultado de la bsqueda");
     TextColor (GRIS);
     Ventana (11,9,73,18);

     RefBD  = CargaBD ( BDCampoReferencia (BD, campo) );
     search = BuscaCampo (RefBD, pk, tema);

     if (search->elementos>0) {

	activo   = InicioLista (search->fichas);

	for ( relativa = 0, aux = InicioLista(search->fichas);
	      aux!=FinLista(search->fichas) && relativa<10;
	      aux=SiguienteLista (search->fichas, aux), relativa++) {
	    ficha = *(long *)LeeNodoLista (search->fichas, aux);
	    ShowRef ( RefBD, pk, ficha, ficha-relativa, (aux==activo));
	}

	relativa = 0;
	ficha    = *(long *)LeeNodoLista (search->fichas, activo);

	do {

	    tecla = GetChar();

	    if (!tecla) {                           /* Cdigos de control */

	       tecla = GetChar();

	       switch (tecla) {

		 case '\x48': /* Arriba */
			      if (activo!=InicioLista(search->fichas)) {
				 ShowRef (RefBD, pk, ficha, ficha-relativa, 0);

				 if (!relativa)
				    ScrollDown (1);
				 else
				    relativa--;

				 activo = AnteriorLista (search->fichas, activo);
				 ficha  = *(long *)LeeNodoLista (search->fichas,activo);
				 ShowRef (RefBD, pk, ficha, ficha-relativa, 1);
			      }
			      break;

		 case '\x50': /* Abajo  */
			      if (activo!=UltimoLista(search->fichas)) {
				 ShowRef (RefBD, pk, ficha, ficha-relativa, 0);

				 if (relativa==9)
				    ScrollUp (1);
				 else
				    relativa++;

				 activo = SiguienteLista (search->fichas, activo);
				 ficha  = *(long *)LeeNodoLista (search->fichas, activo);
				 ShowRef (RefBD, pk, ficha, ficha-relativa, 1);
			      }
			      break;
	       }

	       tecla = '\0';
	    }

	} while (tecla!=ESCAPE && tecla!=RETURN );

	if (tecla!=ESCAPE)
	   *valor = ficha;

     } else {   /* No se ha encontrado el dato buscado */

	GotoXY (32,9);
	TextColor (CYAN);
	Parpadeo (ON);
	PrintF ("Dato no encontrado");
	Parpadeo (OFF);
	TextColor (GRIS);

	do tecla = GetChar();
	while (tecla!=' ' && tecla!=RETURN && tecla!=ESCAPE);

	tecla = ESCAPE;
     }

     DestruyeBusqueda (&search);
     LiberaBD (&RefBD);

  Cursor (cursor);
  Ventana(1,6,80,23);

  return tecla;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

int EditaCampo (Ficha ficha, int campo)
{
  BYTE         x,y;
  int          tecla;
  char         tema[40];
  CommandLine  CL;
  unsigned     cursor = EstadoCursor();

  PosicionCursor (&x,&y);

  switch (Tipo_Campo(ficha->BD, campo)) {

    case Entero:     do {
			GotoXY (x,y);
			tecla = EditaEntero ((long *)Campo(ficha, campo));
		     } while (!CheckCampoEntero(ficha->BD, campo, *(long *)Campo(ficha, campo)));

		     break;

    case Fecha:      do {
			GotoXY (x,y);
			tecla = EditaFecha ((long *)Campo(ficha, campo));
		     } while (!CheckCampoFecha(ficha->BD, campo, *(long *)Campo(ficha, campo)));

		     break;

    case Real:       do {
			GotoXY (x,y);
			tecla = EditaReal ((double *)Campo(ficha, campo));
		     } while (!CheckCampoReal(ficha->BD, campo, *(double*)Campo(ficha,campo)));

		     break;

    case Cadena:     tecla = EditaCadena ( (char*)Campo(ficha,campo),
					  LongitudCampoCadena(ficha->BD, campo) );
		     break;

    case Enumerado:  tecla = EditaEnumerado (ficha->BD, campo, (short *)Campo(ficha,campo) );
		     break;

    case MEMO:       PushPagina();
		     Presentacion (NombreCampo(ficha->BD,campo));
		     Ventana(10,6,71,23);
		     EditaTexto ((Texto)Campo(ficha,campo));
		     Ventana(1,6,80,25);
		     tecla = RETURN;
		     PopPagina();
		     break;

    case Referencia: PushPagina();
		     CL = CreaCL();
		     NuevoComando (CL, "Buscar");

		     do {
			tecla = EditaReferencia (ficha->BD, campo, (long*)Campo(ficha,campo), CL );

			if (tecla=='b' || tecla=='B') {
			   cls();
			   GotoXY (21, 10);
			   TextColor (CYAN);
			   PrintF ("Tema de bsqueda:");
			   GotoXY (21, 12);
			   TextColor (GRIS);
			   tecla = LeeCadena (tema, 40);
			   cls();

			   if (tecla!=ESCAPE) {
			      tecla = BuscaReferencia (ficha->BD, campo, (long*)Campo(ficha,campo), tema);

			      if (tecla==ESCAPE)
				 tecla = 'B';
			   }
			}

		     } while (TeclaCL(CL,tecla));

		     tecla = RETURN;
		     DestruyeCL (&CL);
		     PopPagina();
		     break;
  }

  Cursor (cursor);

  return tecla;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

#define SI 1
#define NO 0

void ShowCampo (Ficha ficha, int campo, int actual)
{
  BYTE x,y;
  char cadena[MAXCADENA];

  PosicionCursor(&x,&y);

  GotoXY (1,y);
  TextColor (CYAN);

  if (campo>1)
     PrintF (SEPARADOR);
  else
     PrintF (CABECERA);

  GotoXY (1,y+1);
  PrintF (INTERMEDIO);
  GotoXY (3,y+1);
  TextColor (GRIS);
  strncpy (cadena,NombreCampo(ficha->BD,campo),CAMPO);
  cadena[CAMPO-1] = '\0';
  PrintF ("%s",cadena);

  if (actual) {
     TextColor (NEGRO);
     TextBackground (GRIS);
  }

  CadenaCampo(ficha,campo,cadena);
  GotoXY (18,y+1);
  PrintF ("%-61s", cadena );

  if (actual) {
     TextColor (GRIS);
     TextBackground (NEGRO);
  }

  TextColor (CYAN);
  GotoXY (1,y+2);

  if (campo==CamposBD(ficha->BD))
     PrintF (PIE);
  else
     PrintF (SEPARADOR);
}

/* ----------------------------------------------------------------------- */

int EditaFicha (Ficha ficha, CommandLine Comandos, char *titulo)
{
  int campo_actual = 1;
  int primer_campo = 1;
  int ultimo_campo ;
  int campo;
  int tecla = ' ';
  int escape;

  unsigned cursor = EstadoCursor();

  Cursor (OCULTO);
  Presentacion (titulo);
  MuestraCL (Comandos, 18);
  Ventana (1,6,80,23);

  for ( campo=1; campo<=CamposBD(ficha->BD) && campo<9; campo++) {
      GotoXY (1,1+(campo-1)*2);
      ShowCampo (ficha, campo, (campo==campo_actual) );
  }
  ultimo_campo = campo-1;

  do {

      if (tecla!=RETURN)
	 tecla = GetChar();

      switch (tecla) {

	case '\x00': tecla = GetChar();

		     switch (tecla) {

		       case '\x48': /* Arriba */
				    if (campo_actual>1) {
				       GotoXY (1, 1+(campo_actual-primer_campo)*2 );
				       ShowCampo (ficha, campo_actual, NO);
				       campo_actual--;

				       if (campo_actual<primer_campo) {
					  primer_campo--;
					  ultimo_campo--;
					  Ventana (1,6,80,22);
					  ScrollDown (2);
					  Ventana (1,6,80,23);
				       }

				       GotoXY (1, 1+(campo_actual-primer_campo)*2 );
				       ShowCampo (ficha, campo_actual, SI);
				    }
				    break;

		       case '\x50': /* Abajo  */
				    if (campo_actual<CamposBD(ficha->BD)) {
				       GotoXY (1, 1+(campo_actual-primer_campo)*2 );
				       ShowCampo (ficha, campo_actual, NO);
				       campo_actual++;

				       if (campo_actual>ultimo_campo) {
					  primer_campo++;
					  ultimo_campo++;
					  Ventana (1,6,80,22);
					  ScrollUp (2);
					  Ventana (1,6,80,23);
				       }

				       GotoXY (1, 1+(campo_actual-primer_campo)*2 );
				       ShowCampo (ficha, campo_actual, SI);
				    }
				    break;
		     }

		     tecla = ' ';
		     break;

	case RETURN: GotoXY (1, 1+(campo_actual-primer_campo)*2 );
		     ShowCampo (ficha, campo_actual, NO);
		     GotoXY (18,2+(campo_actual-primer_campo)*2 );
		     for (campo=0; campo<61; campo++) PutChar(' ');
		     GotoXY (18,2+(campo_actual-primer_campo)*2 );
		     TextColor (GRIS);
		     escape = EditaCampo (ficha, campo_actual);

		     if (  campo_actual<CamposBD(ficha->BD)
			&& escape!=ESCAPE
			&& Tipo_Campo(ficha->BD, campo_actual)!=MEMO
			&& Tipo_Campo(ficha->BD, campo_actual)!=Referencia) {

			GotoXY (1, 1+(campo_actual-primer_campo)*2 );
			ShowCampo (ficha, campo_actual, NO);
			campo_actual++;

			if (campo_actual>ultimo_campo) {
			   primer_campo++;
			   ultimo_campo++;
			   ScrollUp (2);
			   GotoXY (1, 1+(campo_actual-primer_campo)*2 );
			   ShowCampo (ficha, campo_actual, NO);
			}

			if (  Tipo_Campo(ficha->BD,campo_actual)==MEMO
			   || Tipo_Campo(ficha->BD,campo_actual)==Referencia ) {
			   GotoXY (1, 1+(campo_actual-primer_campo)*2 );
			   ShowCampo (ficha, campo_actual, SI);
			   tecla = ' ';
			}

		     } else {
			GotoXY (1, 1+(campo_actual-primer_campo)*2 );
			ShowCampo (ficha, campo_actual, SI);
			tecla = ' ';
		     }

		     break;
      }

  } while (!TeclaCL(Comandos,tecla)&&(tecla!=ESCAPE));

  Cursor (cursor);
  Ventana (1,6,80,23);

  return tecla;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*  Lectura de fichas                                                      */
/* ----------------------------------------------------------------------- */

int LeeFicha ( DataBase BD, Ficha *ficha, CommandLine CL, char *titulo )
{
  /* Crea una ficha nueva */
  *ficha = CreaFicha (BD);
  /* Se edita la ficha recin creada */
  return EditaFicha (*ficha, CL, titulo);
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
