/* ----------------------------------------------------------------------- */
/*         Fernando Berzal Galiano      ETS Ingeniera Informtica         */
/* ----------------------------------------------------------------------- */
/*                   Ordenacin de una tabla en memoria                    */
/* ----------------------------------------------------------------------- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "DB.H"
#include "DB_SORT.H"
#include "DB_FRM.H"

#include "FB_CMP.H"

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*                        Indexacin de una tabla                          */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

static int   dim;
static FCMP  cmp;
static int   offset;
static int   tipo;

static char *buffer;

/* ----------------------------------------------------------------------- */
/*                         Comparacin de fichas                           */
/* ----------------------------------------------------------------------- */

static int CmpFichas ( long f1, long f2 )
{
  void *dato1 = buffer + (f1-1)*dim;
  void *dato2 = buffer + (f2-1)*dim;

  return tipo*cmp(dato1,dato2,offset);
}

/* ----------------------------------------------------------------------- */
/*  Algoritmo de ordenacin: MergeSort iterativo                           */
/* ----------------------------------------------------------------------- */

static void MergeSortDB (long datos[], int N)
{
  long ancho = 2;
  long inicio, indice, izq, der, ext;
  long aux;

  while (ancho<2*N) {

     for (inicio=0; inicio<N-1; inicio+=ancho) {

	 /* Mezcla */

	 izq= inicio;
	 der= inicio + ancho/2;
	 ext= (inicio+ancho<N)? inicio+ancho: N;

	 while (izq<der && der<ext) {

	   while ( CmpFichas(datos[izq],datos[der])<=0 && izq<der)
		 izq++;

	   aux = datos[der];

	   for (indice=der; indice>izq; indice--)
	       datos[indice]=datos[indice-1];

	   datos[izq] = aux;
	   der++;
	   izq++;
	 }
     }

     ancho *= 2;
  }
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

void IndexaTabla (DataBase BD, int campo, char *fich_ndx, int ordenacion)
{
  long  fichas = FichasBD(BD);
  long *orden  = calloc(fichas,sizeof(long));
  FILE *index  = fopen(fich_ndx,"wb");
  FILE *fich   = fopen(BD->fichDatos,"rb");
  long  i;

  dim    = BD->dim;
  offset = OffsetCampo(BD,campo);
  tipo   = ordenacion;
  buffer = calloc (fichas, BD->dim);

  switch (Tipo_Campo(BD,campo)) {

    case Cadena:     cmp = CmpCadena;
		     break;
    case Entero:
    case Fecha:
    case Referencia: cmp = CmpLong;
		     break;

    case Enumerado:  cmp = CmpShort;
		     break;

    case Real:       cmp = CmpDouble;
		     break;

    case MEMO:       cmp = NULL;
		     break;
  }

  if (fich && index && orden && cmp && buffer) {

     for (i=0; i<fichas; i++)
	 orden[i] = i+1;

     fread (buffer, dim, fichas, fich);

     MergeSortDB (orden, fichas);
     fwrite(orden,sizeof(long),fichas,index);
  }

  if (buffer)
     free(buffer);

  if (orden)
     free(orden);

  if (fich)
     fclose(fich);

  if (index)
     fclose(index);
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/*                   Ordenacin fsica de una tabla                        */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

/* ----------------------------------------------------------------------- */
/*                 Reordenacin del fichero principal                      */
/* ----------------------------------------------------------------------- */

static void ReordenaFicheroPrincipal (DataBase BD, long *orden, long fichas)
{
  long  i;
  FILE *fDAT   = fopen(BD->fichDatos,"rb");
  FILE *fdat   = fopen("data.tmp", "wb");
  void *buffer = malloc(BD->dim);

  for (i=0; i<fichas; i++) {
      fseek (fDAT, (orden[i]-1)*(BD->dim), SEEK_SET);
      fread (buffer, BD->dim, 1, fDAT);
      fwrite(buffer, BD->dim, 1, fdat);
  }

  free (buffer);
  fclose (fdat);
  fclose (fDAT);

  remove (BD->fichDatos);
  rename ("data.tmp", BD->fichDatos);
}

/* ----------------------------------------------------------------------- */
/*                         Gestin de campos MEMO                          */
/* ----------------------------------------------------------------------- */

#ifndef MARCA_MEMO
#define MARCA_MEMO ""
#endif

/* ----------------------------------------------------------------------- */

static void LeeCad (FILE *fichero, unsigned char *cadena, int longitud)
{
  fgets (cadena, longitud, fichero);

  if (cadena[strlen(cadena)-1]=='\n')
     cadena[strlen(cadena)-1] = '\0';
  else if (strlen(cadena)>0)
     while (fgetc(fichero)!='\n')
	   ;
}

/* ----------------------------------------------------------------------- */

static void ReordenaMEMO (DataBase BD, int campo, long *orden, long fichas)
{
  long  posicion;
  long  i;
  char  cadena[CAMPOMEMO];
  FILE *fNDX = fopen( IndiceCampoMEMO(BD,campo),"rb");
  FILE *fMEM = fopen(FicheroCampoMEMO(BD,campo),"rb");
  FILE *fmem = fopen("memo.tmp", "wb");
  FILE *fndx = fopen("index.tmp","wb");

  for (i=0; i<fichas; i++) {

      posicion = ftell (fmem);
      fwrite (&posicion, sizeof(long), 1, fndx);

      fseek (fNDX, (orden[i]-1)*sizeof(long), SEEK_SET);
      fread (&posicion, sizeof(long), 1, fNDX);
      fseek (fMEM, posicion, SEEK_SET);

      do {
	  LeeCad (fMEM, cadena, CAMPOMEMO);
	  fprintf(fmem, "%s\n", cadena);
      } while (strcmp(cadena,MARCA_MEMO));

  }

  fclose (fmem);
  fclose (fndx);
  fclose (fMEM);
  fclose (fNDX);

  remove (FicheroCampoMEMO(BD,campo));
  rename ("memo.tmp", FicheroCampoMEMO(BD,campo));
  remove (IndiceCampoMEMO(BD,campo));
  rename ("index.tmp",IndiceCampoMEMO(BD,campo));
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

void OrdenaTabla (DataBase BD, char *fich_ndx)
{
  long  fichas = FichasBD(BD);
  long *orden  = calloc(fichas,sizeof(long));
  FILE *index;
  int   campo;

  if ( orden ) {

     // Fichero ndice (ordenacin deseada)

     index = fopen(fich_ndx,"rb");
     fread (orden,sizeof(long),fichas,index);
     fclose (index);

     // Fichero principal

     ReordenaFicheroPrincipal(BD,orden,fichas);

     // Campos MEMO

     for (campo=1; campo<=BD->NumCampos; campo++)
	 if (Tipo_Campo(BD,campo)==MEMO)
	    ReordenaMEMO (BD,campo,orden,fichas);
  }

  if (orden)
     free(orden);
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

void OrdenarTabla (DataBase BD, int campo, int ordenacion)
{
  long  fichas = FichasBD(BD);
  long *orden  = calloc(fichas,sizeof(long));
  FILE *fich   = fopen(BD->fichDatos,"rb");
  long  i;

  dim    = BD->dim;
  offset = OffsetCampo(BD,campo);
  tipo   = ordenacion;
  buffer = calloc (fichas, BD->dim);

  switch (Tipo_Campo(BD,campo)) {

    case Cadena:     cmp = CmpCadena;
		     break;
    case Entero:
    case Fecha:
    case Referencia: cmp = CmpLong;
		     break;

    case Enumerado:  cmp = CmpShort;
		     break;

    case Real:       cmp = CmpDouble;
		     break;

    case MEMO:       cmp = NULL;
		     break;
  }

  if (fich && orden && cmp && buffer) {

     for (i=0; i<fichas; i++)
	 orden[i] = i+1;

     fread (buffer, dim, fichas, fich);

     // Ordenacin

     MergeSortDB (orden, fichas);

     // Fichero principal

     fclose (fich);
     fich = fopen(BD->fichDatos,"wb");

     for (i=0; i<fichas; i++)
	 fwrite(buffer+(orden[i]-1)*dim, dim, 1, fich);

     // Campos MEMO

     for (campo=1; campo<=BD->NumCampos; campo++)
	 if (Tipo_Campo(BD,campo)==MEMO)
	    ReordenaMEMO (BD,campo,orden,fichas);
  }

  if (buffer)
     free(buffer);

  if (orden)
     free(orden);

  if (fich)
     fclose(fich);
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
