/* ----------------------------------------------------------------------- */
/*         Fernando Berzal Galiano       ETS Ingeniera Informtica        */
/* ----------------------------------------------------------------------- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "DB.H"
#include "DB_SRC.H"
#include "DB_FRM.H"

#include "FB_STR.H"

#include "FBG_DATE.H"

/* ----------------------------------------------------------------------- */
/*                         Gestin de campos MEMO                          */
/* ----------------------------------------------------------------------- */

#ifndef MARCA_MEMO
#define MARCA_MEMO ""
#endif

static void LeeCad (FILE *fichero, unsigned char *cadena, int longitud)
{
  fgets (cadena, longitud, fichero);

  if (cadena[strlen(cadena)-1]=='\n')
     cadena[strlen(cadena)-1] = '\0';
  else
     while (fgetc(fichero)!='\n')
	   ;
}

/* ----------------------------------------------------------------------- */
/*                         Operaciones de bsqueda                         */
/* ----------------------------------------------------------------------- */

void DestruyeBusqueda (Busqueda *b)
{
  DestruyeLista(&((*b)->fichas));
  free (*b);
  *b = NULL;
}

int  Fichas   (Busqueda b)
{
  return b->elementos;
}

long NumFicha (Busqueda b, Nodo n)
{
  return *(long*)LeeNodoLista(b->fichas,n);
}

Lista ListaBusqueda (Busqueda b)
{
  return b->fichas;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

static void BuscaMEMO (char *fichero, Busqueda b, char *tema)
{
  long           ficha = 1;
  long           longitud ;
  unsigned char  cadena[CAMPOMEMO];
  char          *buffer = malloc(BUFFER);
  FILE          *f      = fopen(fichero,"rb");

  if (f) {

     AutomataKMP (tema);

     if (buffer)
	setvbuf (f, buffer, _IOFBF, BUFFER);

     fseek (f, 0, SEEK_END);
     longitud = ftell (f);
     fseek (f, 0, SEEK_SET);

     while ( ftell(f)<longitud ) {

       LeeCad (f, cadena, CAMPOMEMO);

       if (!strcmp(cadena,MARCA_MEMO))
	  ficha++;
       else if (BusquedaKMP(tema, cadena)>=0) {
	  InsertaNodoLista (b->fichas, FinLista(b->fichas), &ficha);
	  b->elementos++;

	  do LeeCad (f, cadena, CAMPOMEMO);
	  while (strcmp(cadena,MARCA_MEMO));

	  ficha++;
       }
     }

     fclose(f);
  }

  if (buffer)
     free (buffer);
}

/* ----------------------------------------------------------------------- */

static void BuscaShort (DataBase BD, int offset, Busqueda b, short tema)
{
  long  ficha  = 1;
  long  fichas ;
  FILE *fich   ;
  void *dato   = malloc (BD->dim);
  char *buffer = malloc (BUFFER);

  if (dato) {

     fich=fopen(BD->fichDatos,"rb");

     if (fich) {

	if (buffer)
	   setvbuf (fich, buffer, _IOFBF, BUFFER);

	fseek (fich, 0, SEEK_END);
	fichas = ftell(fich) / BD->dim;
	fseek (fich, 0, SEEK_SET);

	while (ficha<=fichas) {
	   fread (dato, BD->dim, 1, fich);
	   if (*(short*)((char *)dato+offset)==tema) {
	      InsertaNodoLista (b->fichas, FinLista(b->fichas), &ficha);
	      b->elementos++;
	   }
	   ficha++;
	}

	fclose (fich);
     }

     free (dato);
  }

  if (buffer)
     free (buffer);
}

/* ----------------------------------------------------------------------- */

static void BuscaLong (DataBase BD, int offset, Busqueda b, long tema)
{
  long  ficha  = 1;
  long  fichas ;
  FILE *fich   ;
  void *dato   = malloc(BD->dim);
  char *buffer = malloc (BUFFER);

  if (dato) {

     fich=fopen(BD->fichDatos,"rb");

     if (fich) {

	if (buffer)
	   setvbuf (fich, buffer, _IOFBF, BUFFER);

	fseek (fich, 0, SEEK_END);
	fichas = ftell(fich) / BD->dim;
	fseek (fich, 0, SEEK_SET);

	while (ficha<=fichas) {
	   fread (dato, BD->dim, 1, fich);
	   if (*(long *)((char *)dato+offset)==tema) {
	      InsertaNodoLista (b->fichas, FinLista(b->fichas), &ficha);
	      b->elementos++;
	   }
	   ficha++;
	}

	fclose (fich);
     }

     free (dato);
  }

  if (buffer)
     free (buffer);
}

/* ----------------------------------------------------------------------- */

static void BuscaDouble (DataBase BD, int offset, Busqueda b, double tema)
{
  long  ficha  = 1;
  long  fichas ;
  FILE *fich   ;
  void *dato   = malloc(BD->dim);
  char *buffer = malloc(BUFFER);

  if (dato) {

     fich=fopen(BD->fichDatos,"rb");

     if (fich) {

	if (buffer)
	   setvbuf (fich, buffer, _IOFBF, BUFFER);

	fseek (fich, 0, SEEK_END);
	fichas = ftell(fich) / BD->dim;
	fseek (fich, 0, SEEK_SET);

	while (ficha<=fichas) {
	   fread (dato, BD->dim, 1, fich);
	   if (*(double *)((char *)dato+offset)==tema) {
	      InsertaNodoLista (b->fichas, FinLista(b->fichas), &ficha);
	      b->elementos++;
	   }
	   ficha++;
	}

	fclose (fich);
     }

     free (dato);
  }

  if (buffer)
     free (buffer);
}

/* ----------------------------------------------------------------------- */

static void BuscaCadena (DataBase BD, int offset, Busqueda b, char *tema)
{
  long  ficha  = 1;
  long  fichas ;
  FILE *fich   ;
  void *dato   = malloc (BD->dim);
  char *buffer = malloc (BUFFER);

  AutomataKMP (tema);

  if (dato) {

     fich=fopen(BD->fichDatos,"rb");

     if (fich) {

	if (buffer)
	   setvbuf (fich, buffer, _IOFBF, BUFFER);

	fseek (fich, 0, SEEK_END);
	fichas = ftell(fich) / BD->dim;
	fseek (fich, 0, SEEK_SET);

	while (ficha<=fichas) {
	   fread (dato, BD->dim, 1, fich);
	   if (BusquedaKMP(tema, (char *)dato+offset)>=0) {
	      InsertaNodoLista (b->fichas, FinLista(b->fichas), &ficha);
	      b->elementos++;
	   }
	   ficha++;
	}

	fclose (fich);
     }

     free (dato);
  }

  if (buffer)
     free (buffer);
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

Busqueda BuscaCampo (DataBase BD, int NumCampo, void *tema)
{
  Busqueda b = malloc (sizeof(struct Busqueda));

  if ( b && FichasBD(BD)>0 ) {

     b->fichas    = CreaLista (sizeof(long));
     b->elementos = 0;

     switch (Tipo_Campo(BD,NumCampo)) {

       case Entero:
       case Fecha:
       case Referencia: BuscaLong (BD, OffsetCampo(BD,NumCampo), b, *(long*)tema);
			break;

       case Real:       BuscaDouble (BD, OffsetCampo(BD,NumCampo), b, *(double*)tema);
			break;

       case Cadena:     BuscaCadena (BD, OffsetCampo(BD,NumCampo), b, (char*)tema);
			break;

       case Enumerado:  BuscaShort (BD, OffsetCampo(BD,NumCampo), b, *(short*)tema);
			break;

       case MEMO:       BuscaMEMO (FicheroCampoMEMO(BD,NumCampo), b, (char*)tema);
			break;
     }

  } else if (b) {
     free (b);
     b = NULL;
  }

  return b;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
