/* ----------------------------------------------------------------------- */
/*         Fernando Berzal Galiano       ETS Ingeniera Informtica        */
/* ----------------------------------------------------------------------- */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include "FBG_CRT.H"
#include "FBG_KBD.H"

#include "FBG_MENU.H"

/* ----------------------------------------------------------------------- */

Menu CreaMenu (void)
{
  Menu  menu = malloc(sizeof(InfoMenu));

  if (menu) {

     menu->elementos = 0;
     menu->info      = CreaLista (sizeof(MenuItem));

     if (!(menu->info)) {
	free (menu);
	menu = NULL;
     }

  }

  return menu;
}

/* ----------------------------------------------------------------------- */

void OpcionMenu (Menu menu, char tecla, char *nombre)
{
  MenuItem item;

  item.tecla = toupper(tecla);
  strncpy (item.nombre, nombre, MENU);

  InsertaNodoLista (menu->info, FinLista(menu->info), &item);
  menu->elementos++;
}

/* ----------------------------------------------------------------------- */

void DestruyeMenu (Menu *menu)
{
  DestruyeLista (&((*menu)->info));
  free (*menu);
  *menu = NULL;
}

/* ----------------------------------------------------------------------- */

#define MAXIMO(a,b) (((a)>(b))?(a):(b))

#define OPCIONES_PANTALLA 18

int EjecutaMenu (Menu menu)
{
  int Espaciado = (menu->elementos>9)? 1: 2;
  int X = 40 - (MENU + 8)/2;
  int InicY = MAXIMO (10 - Espaciado*(menu->elementos)/2, 1);
  int Y;
  int OpcionResaltada = 1;
  int OpcionInicial   = 1;
  int tecla;
  int Encontrado = 0;
  int OK = 0;
  int i, j;
  unsigned cursor = EstadoCursor();

  Nodo      nodo;
  Nodo      inicio = InicioLista(menu->info);
  MenuItem *item;

  Cursor (OCULTO);
  cls();

  while (!OK) {

     /* Muestra men en pantalla */

     for ( i=0, Y=InicY, nodo=inicio;
	   nodo!=FinLista(menu->info) && i<OPCIONES_PANTALLA;
	   i++, Y+=Espaciado, nodo=SiguienteLista(menu->info,nodo) ) {

	 GotoXY (X, Y);
	 item = LeeNodoLista (menu->info, nodo);

	 if (OpcionResaltada==OpcionInicial+i)
	    TextBackground (AZUL);

	 PrintF (" %c ", item->tecla);
	 for (j=1; j<=(MENU+3-strlen(item->nombre)); j++) PrintF(".");
	 PrintF (" %s ", item->nombre);

	 if (OpcionResaltada==OpcionInicial+i)
	    TextBackground (NEGRO);
     }

     /* Lee opcin elegida */

     tecla=GetChar();

     if (!tecla) {                              /* Cdigos de control '\0' */

	tecla = GetChar();

	switch (tecla) {

	  case '\x48': if (OpcionResaltada>1) {               /* Arriba    */
			  OpcionResaltada--;

			  if (OpcionResaltada<OpcionInicial) {
			     inicio = AnteriorLista(menu->info,inicio);
			     OpcionInicial--;
			  }
		       }
		       break;

	  case '\x50': if (OpcionResaltada<menu->elementos) { /* Abajo     */
			  OpcionResaltada++;

			  if (OpcionResaltada-OpcionInicial>=OPCIONES_PANTALLA) {
			     inicio = SiguienteLista(menu->info,inicio);
			     OpcionInicial++;
			  }
		       }
		       break;

	  case '\x4B':                                        /* Izquierda */
	  case '\x47': OpcionResaltada = 1;                   /* Inicio    */
		       inicio = InicioLista(menu->info);
		       OpcionInicial = 1;
		       break;

	  case '\x4D':                                        /* Derecha   */
	  case '\x4F': OpcionResaltada = menu->elementos;     /* Fin       */

		       if (OpcionResaltada-OpcionInicial>=OPCIONES_PANTALLA) {
			  inicio = UltimoLista(menu->info);
			  OpcionInicial = menu->elementos;
			  for ( i=1; i<OPCIONES_PANTALLA; i++) {
			      inicio = AnteriorLista(menu->info,inicio);  
			      OpcionInicial--;
			  }
		       }
		       break;

	}

     } else if (tecla==RETURN) {        /* Eleccin de la opcin resaltada */

	nodo = InicioLista (menu->info);

	for (i=1; i<OpcionResaltada; i++)
	    nodo = SiguienteLista (menu->info, nodo);

	item  = LeeNodoLista (menu->info, nodo);
	tecla = item->tecla;
	OK = 1;

     } else if (tecla==ESCAPE) {

	OK = 1;

     } else {

	nodo = InicioLista(menu->info);

	for (i=1; (i<=menu->elementos)&&(!Encontrado); i++) {  /* Hot keys */

	  item = (MenuItem *)LeeNodoLista(menu->info, nodo);
	  nodo = SiguienteLista (menu->info, nodo);

	  if ( item->tecla==toupper(tecla) ) {
	     OpcionResaltada = i;
	     Encontrado = 1;

	     if (OpcionResaltada<OpcionInicial) {

		while (OpcionResaltada<OpcionInicial) {
		      OpcionInicial--;
		      inicio = AnteriorLista(menu->info,inicio);
		}

	     } else if (OpcionResaltada-OpcionInicial>=OPCIONES_PANTALLA) {

		while (OpcionResaltada-OpcionInicial>=OPCIONES_PANTALLA) {
		      OpcionInicial++;
		      inicio = SiguienteLista(menu->info,inicio);
		}

	     }
	  }
	}

	Encontrado = 0;

     }

  }

  Cursor (cursor);
  return tecla;
}

/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */

CommandLine CreaCL (void)
{
  CommandLine CL = malloc(sizeof(InfoCL));

  if (CL) {

     CL->elementos = 0;
     CL->longitud  = 0;
     CL->comandos  = CreaLista (sizeof(Comando));

     if (!(CL->comandos)) {
	free (CL);
	CL = NULL;
     }
  }

  return CL;
}

/* ----------------------------------------------------------------------- */

void NuevoComando (CommandLine CL, char *nombre)
{
 Nodo     nodo  = InicioLista (CL->comandos);
 Comando *aux;
 Comando  c;

 c.letra = 0;

 while (nodo!=FinLista(CL->comandos)) {

    aux = LeeNodoLista (CL->comandos, nodo);

    if (toupper(nombre[c.letra])==aux->nombre[aux->letra]) {
       c.letra++;
       nodo = InicioLista (CL->comandos);
    } else
       nodo = SiguienteLista (CL->comandos, nodo);
 }

 strncpy (c.nombre, nombre, COMANDO);
 InsertaNodoLista (CL->comandos, FinLista(CL->comandos), &c);
 CL->elementos++;

 if (CL->elementos==1)
    CL->longitud = strlen(c.nombre);
 else
    CL->longitud += strlen(c.nombre)+2;
}

/* ----------------------------------------------------------------------- */

void DestruyeCL (CommandLine *CL)
{
  DestruyeLista (&((*CL)->comandos));
  free (*CL);
  *CL = NULL;
}

/* ----------------------------------------------------------------------- */

void MuestraCL (CommandLine CL, int linea)
{
  int      i;
  Comando *c;
  Nodo     nodo;

  GotoXY (41-(CL->longitud)/2, linea);

  for ( nodo = InicioLista(CL->comandos);
	nodo!= FinLista(CL->comandos);
	nodo = SiguienteLista(CL->comandos, nodo) ) {

    c = LeeNodoLista (CL->comandos, nodo);

    for (i=0; i<strlen(c->nombre); i++) {
	if (i==c->letra) TextColor (CYAN);
	PrintF ("%c",c->nombre[i]);
	if (i==c->letra) TextColor (GRIS);
    }

    if (nodo!=UltimoLista(CL->comandos))
       PrintF("  ");
  }
}

/* ----------------------------------------------------------------------- */

int TeclaCL (CommandLine CL, int tecla)
{
  int      Encontrado=0;
  Comando *c;
  Nodo     nodo = InicioLista (CL->comandos);

  while ( nodo!=FinLista(CL->comandos) && !Encontrado ) {

    c    = LeeNodoLista (CL->comandos, nodo);
    nodo = SiguienteLista (CL->comandos, nodo);

    if (toupper(c->nombre[c->letra])==toupper(tecla))
       Encontrado = 1;
  }

  return Encontrado;
}

/* ----------------------------------------------------------------------- */
/*  YES-NO BOX                                                             */
/* ----------------------------------------------------------------------- */

void MuestraYN (int YN)
{
  if (YN)
     TextColor (BLANCO);
  else
     TextColor (NEGRO);

  GotoXY (38,4);
  PrintF ("S");

  if (!YN)
     TextColor (BLANCO);
  else
     TextColor (NEGRO);

  GotoXY (42,4);
  PrintF ("No");
}


int YesNo (const char *cadena, int YN)
{
  int      x      = strlen(cadena)/2 + 2;
  int      tecla  ;
  unsigned cursor = EstadoCursor ();

  Cursor (OCULTO);

  while (TeclaPulsada()) GetChar();

  TextBackground (AZUL);
  Ventana ( 41-x, 19, 40+x+(strlen(cadena)%2?1:0), 23);
  cls ();
  Ventana ( 1, 19, 80, 23);
  GotoXY ( 43-x, 2);
  PrintF("%s", cadena);

  do {
      MuestraYN (YN);
      tecla = GetChar();

      switch (tecla) {

	case 0: /* Cdigos de control */
		tecla = GetChar();

		switch (tecla) {

		  case '\x4B': /* Izquierda */
		  case '\x48': /* Arriba    */
		  case '\x4D': /* Derecha   */
		  case '\x50': /* Abajo     */
			       YN = !YN;
			       break;
		}

		tecla = ' ';
		break;

	case ESCAPE:
		YN = 0;
		break;
      }

  } while (tecla!=RETURN);

  TextBackground (NEGRO);
  TextColor (GRIS);
  cls ();

  Ventana (  1,  6, 80, 23 );

  Cursor (cursor);

  return YN;
}
